import concurrent.futures
import multiprocessing as mp
import pyarrow
import pyarrow.dataset as pyarrow_dataset # need to import separately, it's not on the pyarrow import
import tempfile
import uuid
import os
import pathlib
from ..classes.Row import Row
from tqdm.auto import tqdm
import shutil
from ..common.api_request import make_request

class RedivisArrowIterator:
    def __init__(self, streams, mapped_variables, progressbar, coerce_schema):
        self.streams = streams
        self.mapped_variables = mapped_variables
        self.progressbar = progressbar
        self.coerce_schema = coerce_schema
        self.current_stream_index = 0
        self.__get_next_reader__()

    def __get_next_reader__(self):
        arrow_response = make_request(
            method="get",
            path=f'/readStreams/{self.streams[self.current_stream_index]["id"]}',
            stream=True,
            parse_response=False,
        )
        self.current_record_batch_reader = pyarrow.ipc.RecordBatchStreamReader(arrow_response.raw)
        if self.coerce_schema:
            self.variables_in_stream = list(
                map(lambda field_name: next(x for x in self.mapped_variables if x["name"] == field_name),
                    self.current_record_batch_reader.schema.names))
            self.output_schema = pyarrow.schema(map(variable_to_field, self.variables_in_stream))
        else:
            self.output_schema = self.current_record_batch_reader.schema

    def __iter__(self):
        return self

    def __next__(self):
        try:
            batch = self.current_record_batch_reader.read_next_batch()
            if self.coerce_schema:
                batch = pyarrow.RecordBatch.from_arrays(
                    list(map(coerce_string_variable, batch.columns, self.variables_in_stream)), schema=self.output_schema)

            if self.progressbar is not None:
                self.progressbar.update(batch.num_rows)

            return batch
        except StopIteration:
            if self.current_stream_index == len(self.streams) - 1:
                if self.progressbar:
                    self.progressbar.close()
                raise StopIteration
            else:
                self.current_stream_index += 1
                self.__get_next_reader__()
                return self.__next__()

def list_rows(
    *,
    uri,
    output_type="dataframe",
    max_results=None,
    selected_variables=None,
    mapped_variables=None,
    target_parallelization=None,
    progress=True,
    coerce_schema = False,
    batch_preprocessor = None
):
    if target_parallelization is None:
        target_parallelization = mp.cpu_count()

    read_session = make_request(
        method="post",
        path=f'{uri}/readSessions',
        parse_response=True,
        payload={
            "selectedVariables": selected_variables,
            "maxResults": max_results,
            "format": "arrow",
            "requestedStreamCount": min(target_parallelization, 16)
        },
    )

    folder = f'/{tempfile.gettempdir()}/redivis/tables/{uuid.uuid4()}'
    pathlib.Path(folder).mkdir(parents=True, exist_ok=True)

    progressbar = None
    if progress:
        progressbar = tqdm(total=max_results, leave=False)

    if output_type == 'arrow_iterator':
        return RedivisArrowIterator(streams=read_session["streams"], mapped_variables=mapped_variables, progressbar=progressbar, coerce_schema=coerce_schema )


    # Use download_state to notify worker threads when to quit.
    # See: https://stackoverflow.com/a/29237343/101923
    download_state = {"done": False}

    # Code to use multiprocess... would simplify exiting on stop, but progress doesn't currently work
    # with concurrent.futures.ProcessPoolExecutor(max_workers=len(read_session["streams"]), mp_context=mp.get_context('fork')) as executor:

    # See https://github.com/googleapis/python-bigquery/blob/main/google/cloud/bigquery/_pandas_helpers.py#L920
    with concurrent.futures.ThreadPoolExecutor(max_workers=len(read_session["streams"])) as executor:
        futures = [executor.submit(process_stream, stream, folder, mapped_variables, coerce_schema, progressbar, download_state, batch_preprocessor)
                   for stream in read_session["streams"]]

        not_done = futures

        try:
            while not_done:
                # next line 'sleeps' this main thread, letting the thread pool run
                freshly_done, not_done = concurrent.futures.wait(not_done, timeout=0.2)
                for future in freshly_done:
                    # Call result() on any finished threads to raise any exceptions encountered.
                    future.result()
        finally:
            for future in not_done:
                # Only cancels futures that were never started
                future.cancel()
            download_state["done"] = True
            # Shutdown all background threads, now that they should know to exit early.
            executor.shutdown(wait=True, cancel_futures=True)

    if progress:
        progressbar.close()

    arrow_dataset = pyarrow_dataset.dataset(folder, format="feather", schema = pyarrow.schema(map(variable_to_field, mapped_variables)) if batch_preprocessor is None else None)

    if output_type == 'arrow_dataset':
        return arrow_dataset
    elif output_type == 'polars_lazyframe':
        import polars
        return polars.scan_ipc(f'{folder}/*', memory_map=True)
    elif output_type == 'dask_dataframe':
        import dask.dataframe as dd
        # TODO: simplify once dask supports reading from feather: https://github.com/dask/dask/issues/6865
        parquet_base_dir = f'/{tempfile.gettempdir()}/redivis/tables/{uuid.uuid4()}'
        pyarrow_dataset.write_dataset(arrow_dataset, parquet_base_dir, format='parquet')
        shutil.rmtree(folder)
        return dd.read_parquet(parquet_base_dir, dtype_backend='pyarrow')
    else:
        # TODO: remove head() once BE is sorted, instead use dataset.to_table()
        arrow_table = pyarrow_dataset.Scanner.from_dataset(arrow_dataset).head(max_results)
        shutil.rmtree(folder)
        if output_type == 'arrow_table':
            return arrow_table
        elif output_type == 'tuple':
            variable_name_to_index = {}
            for index, variable in enumerate(mapped_variables):
                variable_name_to_index[variable["name"]] = index

            pydict = arrow_table.to_pydict()
            keys = list(pydict.keys())

            return [
                Row([pydict[variable["name"]][i] for variable in mapped_variables], variable_name_to_index)
                for i in range(len(pydict[keys[0]]))
            ]


def variable_to_field(variable):
    if variable["type"] == 'string' or variable["type"] == 'geography':
        return pyarrow.field(variable["name"], pyarrow.string())
    elif variable["type"] == 'integer':
        return pyarrow.field(variable["name"], pyarrow.int64())
    elif variable["type"] == 'float':
        return pyarrow.field(variable["name"], pyarrow.float64())
    elif variable["type"] == 'date':
        return pyarrow.field(variable["name"], pyarrow.date32())
    elif variable["type"] == 'dateTime':
        return pyarrow.field(variable["name"], pyarrow.timestamp('us'))
    elif variable["type"] == 'time':
        return pyarrow.field(variable["name"], pyarrow.time64('us'))
    elif variable["type"] == 'boolean':
        return pyarrow.field(variable["name"], pyarrow.bool_())


def coerce_string_variable(pyarrow_array, variable):
    if variable["type"] == 'string' or variable["type"] == 'geography':
        return pyarrow_array
    elif variable["type"] == 'integer':
        return pyarrow.compute.cast(pyarrow_array, pyarrow.int64())
    elif variable["type"] == 'float':
        return pyarrow.compute.cast(pyarrow_array, pyarrow.float64())
    elif variable["type"] == 'date':
        return pyarrow.compute.cast(pyarrow.compute.cast(pyarrow_array, pyarrow.timestamp('us')), pyarrow.date32())
    elif variable["type"] == 'dateTime':
        return pyarrow.compute.cast(pyarrow_array, pyarrow.timestamp('us'))
    elif variable["type"] == 'time':
        return pyarrow.compute.cast(
            pyarrow.compute.cast(
                pyarrow.compute.utf8_replace_slice(pyarrow_array,start=0,stop=0,replacement='2020-01-01T'),
                pyarrow.timestamp('us')
            ),
            pyarrow.time64('us')
        )
    elif variable["type"] == 'boolean':
        return pyarrow.compute.cast(pyarrow_array, pyarrow.bool_())


def process_stream(stream, folder, mapped_variables, coerce_schema, progressbar, download_state, batch_preprocessor):
    arrow_response = make_request(
        method="get",
        path=f'/readStreams/{stream["id"]}',
        stream=True,
        parse_response=False,
    )

    reader = pyarrow.ipc.RecordBatchStreamReader(arrow_response.raw)

    if coerce_schema:
        variables_in_stream = list(map(lambda field_name: next(x for x in mapped_variables if x["name"] == field_name), reader.schema.names))
        output_schema = pyarrow.schema(map(variable_to_field, variables_in_stream))
    else:
        output_schema = reader.schema

    has_content = False
    with pyarrow.OSFile(f"{folder}/{stream['id']}", mode="wb") as f:
        writer = None
        for batch in reader:
            # exit out of thread
            if download_state["done"]:
                break

            if coerce_schema:
                batch = pyarrow.RecordBatch.from_arrays(list(map(coerce_string_variable, batch.columns, variables_in_stream)), schema=output_schema)

            num_rows = batch.num_rows
            if batch_preprocessor:
                batch = batch_preprocessor(batch)

            if batch is not None:
                has_content = True
                if writer is None:
                    writer = pyarrow.ipc.RecordBatchFileWriter(f, output_schema if batch_preprocessor is None else batch.schema)

                writer.write_batch(batch)

            if progressbar is not None:
                progressbar.update(num_rows)

        if writer is not None:
            writer.close()

    if has_content == False:
        os.remove(f"{folder}/{stream['id']}")

    reader.close()


def format_tuple_type(val, type):
    if val is None:
        return val
    elif type == "integer":
        return int(val)
    elif type == "float":
        return float(val)
    elif type == "date":
        return str(val)
    elif type == "dateTime":
        return str(val)
    elif type == "time":
        return str(val)
    elif type == "boolean":
        return bool(val)
    else:
        return str(val)





from .classes.User import User as user
from .classes.Organization import Organization as organization
from .classes.Query import Query as query
from .classes.Table import Table as table
from .classes.File import File as file

from ._version import __version__


def current_notebook():
    import os
    if os.getenv("REDIVIS_NOTEBOOK_JOB_ID") is not None:
        from .classes.Notebook import Notebook
        return Notebook(os.getenv("REDIVIS_NOTEBOOK_JOB_ID"))

import pkg_resources  # part of setuptools



from .Base import Base
import os
import time
import warnings
import pyarrow as pa
import pandas as pd

from ..common.api_request import make_request
from ..common.list_rows import list_rows
from ..common.util import get_geography_variable, get_warning


class Query(Base):
    def __init__(
        self,
        query,
        *,
        default_project=None,
        default_dataset=None,
    ):
        if not default_project and not default_dataset:
            if os.getenv("REDIVIS_DEFAULT_PROJECT"):
                default_project = os.getenv("REDIVIS_DEFAULT_PROJECT")
            elif os.getenv("REDIVIS_DEFAULT_DATASET"):
                default_dataset = os.getenv("REDIVIS_DEFAULT_DATASET")

        self.properties = make_request(
            method="post",
            path="/queries",
            payload={
                "query": query,
                "defaultProject": default_project if default_project else None,
                "defaultDataset": default_dataset if default_dataset else None,
            },
        )
        self.uri = f"/queries/{self.properties['id']}"

    def get(self):
        self.properties = make_request(method="GET", path=self.uri)
        return self

    def to_arrow_dataset(self, max_results=None, *, progress=True, batch_preprocessor=None):
        self._wait_for_finish()

        return list_rows(
            uri=self.uri,
            max_results=self.properties["outputNumRows"] if max_results is None else min(max_results, int(
                self.properties["outputNumRows"])),
            mapped_variables=self.properties["outputSchema"],
            output_type="arrow_dataset",
            progress=progress,
            coerce_schema=False,
            batch_preprocessor=batch_preprocessor
        )

    def to_arrow_table(self, max_results=None, *, progress=True, batch_preprocessor=None):
        self._wait_for_finish()

        return list_rows(
            uri=self.uri,
            max_results=self.properties["outputNumRows"] if max_results is None else min(max_results, int(
                self.properties["outputNumRows"])),
            mapped_variables=self.properties["outputSchema"],
            output_type="arrow_table",
            progress=progress,
            coerce_schema=False,
            batch_preprocessor=batch_preprocessor
        )

    def to_polars_lazyframe(self, max_results=None, *, progress=True, batch_preprocessor=None):
        self._wait_for_finish()

        return list_rows(
            uri=self.uri,
            max_results=self.properties["outputNumRows"] if max_results is None else min(max_results, int(
                self.properties["outputNumRows"])),
            mapped_variables=self.properties["outputSchema"],
            output_type="polars_lazyframe",
            progress=progress,
            coerce_schema=False,
            batch_preprocessor=batch_preprocessor
        )

    def to_dask_dataframe(self, max_results=None, *, progress=True, batch_preprocessor=None):
        self._wait_for_finish()

        return list_rows(
            uri=self.uri,
            max_results=self.properties["outputNumRows"] if max_results is None else min(max_results, int(
                self.properties["outputNumRows"])),
            mapped_variables=self.properties["outputSchema"],
            output_type="dask_dataframe",
            progress=progress,
            coerce_schema=False,
            batch_preprocessor=batch_preprocessor
        )

    def to_pandas_dataframe(self, max_results=None, *, geography_variable="", progress=True, dtype_backend='pyarrow', date_as_object=False, batch_preprocessor=None):
        if dtype_backend not in ['numpy', 'numpy_nullable', 'pyarrow']:
            raise Exception(f"Unknown dtype_backend. Must be one of 'pyarrow'|'numpy_nullable'|'numpy'. Default is 'pyarrow'")

        self._wait_for_finish()

        arrow_table = list_rows(
            uri=self.uri,
            max_results=self.properties["outputNumRows"] if max_results is None else min(max_results, int(self.properties["outputNumRows"])),
            mapped_variables=self.properties["outputSchema"],
            output_type="arrow_table",
            progress=progress,
            coerce_schema=False,
            batch_preprocessor=batch_preprocessor
        )

        if dtype_backend == 'numpy_nullable':
            df = arrow_table.to_pandas(self_destruct=True, date_as_object=date_as_object, types_mapper={
                pa.int64(): pd.Int64Dtype(),
                pa.bool_(): pd.BooleanDtype(),
                pa.float64(): pd.Float64Dtype(),
                pa.string(): pd.StringDtype(),
            }.get)
        elif dtype_backend == 'pyarrow':
            df = arrow_table.to_pandas(self_destruct=True, types_mapper=pd.ArrowDtype)
        else:
            df = arrow_table.to_pandas(self_destruct=True, date_as_object=date_as_object)

        if geography_variable is not None:
            geography_variable = get_geography_variable(self.properties["outputSchema"], geography_variable)

        if geography_variable is not None:
            import geopandas
            df[geography_variable["name"]] = geopandas.GeoSeries.from_wkt(df[geography_variable["name"]])
            df = geopandas.GeoDataFrame(data=df, geometry=geography_variable["name"], crs="EPSG:4326")

        return df

    def to_geopandas_dataframe(self, max_results=None, *, geography_variable="", progress=True, dtype_backend='pyarrow', date_as_object=False, batch_preprocessor=None):
        import geopandas

        if dtype_backend not in ['numpy', 'numpy_nullable', 'pyarrow']:
            raise Exception(f"Unknown dtype_backend. Must be one of 'pyarrow'|'numpy_nullable'|'numpy'. Default is 'pyarrow'")

        self._wait_for_finish()

        arrow_table = list_rows(
            uri=self.uri,
            max_results=self.properties["outputNumRows"] if max_results is None else min(max_results, int(self.properties["outputNumRows"])),
            mapped_variables=self.properties["outputSchema"],
            output_type="arrow_table",
            progress=progress,
            coerce_schema=False,
            batch_preprocessor=batch_preprocessor
        )

        if dtype_backend == 'numpy_nullable':
            df = arrow_table.to_pandas(self_destruct=True, date_as_object=date_as_object, types_mapper={
                pa.int64(): pd.Int64Dtype(),
                pa.bool_(): pd.BooleanDtype(),
                pa.float64(): pd.Float64Dtype(),
                pa.string(): pd.StringDtype(),
            }.get)
        elif dtype_backend == 'pyarrow':
            df = arrow_table.to_pandas(self_destruct=True, types_mapper=pd.ArrowDtype)
        else:
            df = arrow_table.to_pandas(self_destruct=True, date_as_object=date_as_object)

        if geography_variable is not None:
            geography_variable = get_geography_variable(self.properties["outputSchema"], geography_variable)
            if geography_variable is None:
                raise Exception('Unable to find a variable with type=="geography" in the query results')

        df[geography_variable["name"]] = geopandas.GeoSeries.from_wkt(df[geography_variable["name"]])
        df = geopandas.GeoDataFrame(data=df, geometry=geography_variable["name"], crs="EPSG:4326")

        return df

    def to_dataframe(self, max_results=None, *, geography_variable="", progress=True):
        warnings.warn(get_warning('dataframe_deprecation'), FutureWarning, stacklevel=2)
        self._wait_for_finish()

        arrow_table = list_rows(
            uri=self.uri,
            max_results=self.properties["outputNumRows"] if max_results is None else min(max_results, int(
                self.properties["outputNumRows"])),
            mapped_variables=self.properties["outputSchema"],
            output_type="arrow_table",
            progress=progress,
            coerce_schema=False
        )

        df = arrow_table.to_pandas(self_destruct=True)

        if geography_variable is not None:
            geography_variable = get_geography_variable(self.properties["outputSchema"], geography_variable)

        if geography_variable is not None:
            import geopandas
            warnings.warn(get_warning('geodataframe_deprecation'), FutureWarning, stacklevel=2)
            df[geography_variable["name"]] = geopandas.GeoSeries.from_wkt(df[geography_variable["name"]])
            df = geopandas.GeoDataFrame(data=df, geometry=geography_variable["name"], crs="EPSG:4326")

        return df

    def to_arrow_batch_iterator(self, max_results=None, *, variables=None, progress=True):
        self._wait_for_finish()

        return list_rows(
            uri=self.uri,
            max_results=self.properties["numRows"] if max_results is None else min(max_results, int(self.properties["numRows"])),
            selected_variables=variables,
            mapped_variables=self.properties["outputSchema"],
            output_type="arrow_iterator",
            progress=progress,
            coerce_schema=hasattr(self.properties, "container") is False or self.properties["container"][
                "kind"] == 'dataset'
        )

    def list_rows(self, max_results=None, *, progress=True):
        warnings.warn(
            "The list_rows method is deprecated. Please use table.to_arrow_batch_iterator() or table.to_arrow_table().to_pylist() for better performance and memory utilization.",
            FutureWarning, stacklevel=2)

        self._wait_for_finish()

        return list_rows(
            uri=self.uri,
            max_results=self.properties["outputNumRows"] if max_results is None else min(max_results, int(
                self.properties["outputNumRows"])),
            mapped_variables=self.properties["outputSchema"],
            output_type="tuple",
            progress=progress,
            coerce_schema=False
        )

    def _wait_for_finish(self):
        while True:
            if self.properties["status"] == "completed":
                break
            elif self.properties["status"] == "failed":
                raise Exception(
                    f"Query job failed with message: {self.properties['errorMessage']}"
                )
            elif self.properties["status"] == "cancelled":
                raise Exception(f"Query job was cancelled")
            else:
                time.sleep(2)
                self.get()



import time
import re
import math
import os
import json
import logging
import requests
from urllib.parse import quote as quote_uri
import warnings
import io
import pyarrow as pa
import pandas as pd

from ..common.util import get_geography_variable, get_warning
from ..common.list_rows import list_rows

from .Base import Base
from .Variable import Variable
from ..common.api_request import make_request, make_paginated_request

# 8MB
MAX_CHUNK_SIZE = 2 ** 23


class Upload(Base):
    def __init__(
        self,
        name,
        *,
        table,
        properties={},
    ):
        self.table = table
        self.name = name
        self.uri = f"{table.uri}/uploads/{quote_uri(self.name,'')}"
        self.properties = {
            **{
                "kind": "upload",
                "name": name,
                "uri": self.uri
            },
            **properties
        }

    def create(
        self,
        data=None,
        *,
        type=None,
        transfer_specification=None,
        delimiter=None,
        schema=None,
        metadata=None,
        has_header_row=True,
        skip_bad_records=False,
        has_quoted_newlines=None,
        quote_character=None,
        rename_on_conflict=False,
        replace_on_conflict=False,
        allow_jagged_rows=False,
        if_not_exists=False,
        remove_on_fail=False,
        wait_for_finish=True,
        raise_on_fail=True,
    ):
        resumable_upload_id = None

        if data and (
            (hasattr(data, "read") and os.stat(data.name).st_size > 1e7)
            or (hasattr(data, "__len__") and len(data) > 1e7)
        ):
            resumable_upload_id = upload_file(data, self.table.uri)
            data = None
        elif data and not hasattr(data, "read"):
            data = io.StringIO(data)

        if schema and type != "stream":
            warnings.warn("The schema option is ignored for uploads that aren't of type `stream`")

        exists = self.exists()

        if if_not_exists and exists:
            return self

        if replace_on_conflict is True and rename_on_conflict is True:
            raise Exception("Invalid parameters. replace_on_conflict and rename_on_conflict cannot both be True.")

        if exists:
            if replace_on_conflict is True:
                self.delete()
            elif rename_on_conflict is not True:
                raise Exception(f"An upload with the name {self.name} already exists on this version of the table. If you want to upload this file anyway, set the parameter rename_on_conflict=True or replace_on_conflict=True.")

        files = None
        payload = {
            "name": self.name,
            "type": type,
            "schema": schema,
            "metadata": metadata,
            "hasHeaderRow": has_header_row,
            "skipBadRecords": skip_bad_records,
            "hasQuotedNewlines": has_quoted_newlines,
            "allowJaggedRows": allow_jagged_rows,
            "quoteCharacter": quote_character,
            "delimiter": delimiter,
            "resumableUploadId": resumable_upload_id,
            "transferSpecification": transfer_specification
        }

        if data is not None:
            files={'metadata': json.dumps(payload), 'data': data}
            payload=None

        response = make_request(
            method="POST",
            path=f"{self.table.uri}/uploads",
            parse_payload=data is None,
            payload=payload,
            files=files
        )
        self.properties = response
        self.uri = self.properties["uri"]

        try:
            if (data or resumable_upload_id or transfer_specification) and wait_for_finish:
                while True:
                    time.sleep(2)
                    self.get()
                    if self["status"] == "completed" or self["status"] == "failed":
                        if self["status"] == "failed" and raise_on_fail:
                            raise Exception(self["errorMessage"])
                        break
                    else:
                        logging.debug("Upload is still in progress...")
        except Exception as e:
            if remove_on_fail:
                self.delete()
            raise Exception(str(e))

        return self

    def delete(self):
        make_request(
            method="DELETE",
            path=self.uri,
        )

    def exists(self):
        try:
            make_request(method="GET", path=self.uri)
            return True
        except Exception as err:
            if err.args[0]["status"] != 404:
                raise err
            return False

    def get(self):
        self.properties = make_request(
            method="GET",
            path=self.uri,
        )
        self.uri = self.properties["uri"]

    def insert_rows(self, rows, *, update_schema=False):
        response = make_request(
            method="POST",
            path=f"{self.uri}/rows",
            payload={"rows": rows, "updateSchema": update_schema},
        )
        return response

    def list_variables(self, *, max_results=10000):
        variables = make_paginated_request(
            path=f"{self.uri}/variables", page_size=1000, max_results=max_results
        )
        return [
            Variable(variable["name"], properties=variable, upload=self)
            for variable in variables
        ]

    def to_arrow_dataset(self, max_results=None, *, variables=None, progress=True, batch_preprocessor=None):
        if not self.properties or not hasattr(self.properties, "numRows"):
            self.get()

        mapped_variables = get_mapped_variables(variables, self.uri)

        return list_rows(
            uri=self.uri,
            max_results=self.properties["numRows"] if max_results is None else min(max_results,
                                                                                   int(self.properties["numRows"])),
            selected_variables=variables,
            mapped_variables=mapped_variables,
            output_type="arrow_dataset",
            progress=progress,
            coerce_schema=hasattr(self.properties, "container") is False or self.properties["container"][
                "kind"] == 'dataset',
            batch_preprocessor=batch_preprocessor
        )

    def to_arrow_table(self, max_results=None, *, variables=None, progress=True, batch_preprocessor=None):
        if not self.properties or not hasattr(self.properties, "numRows"):
            self.get()

        mapped_variables = get_mapped_variables(variables, self.uri)

        return list_rows(
            uri=self.uri,
            max_results=self.properties["numRows"] if max_results is None else min(max_results,
                                                                                   int(self.properties["numRows"])),
            selected_variables=variables,
            mapped_variables=mapped_variables,
            output_type="arrow_table",
            progress=progress,
            coerce_schema=hasattr(self.properties, "container") is False or self.properties["container"][
                "kind"] == 'dataset',
            batch_preprocessor=batch_preprocessor
        )

    def to_polars_lazyframe(self, max_results=None, *, variables=None, progress=True, batch_preprocessor=None):
        if not self.properties or not hasattr(self.properties, "numRows"):
            self.get()

        mapped_variables = get_mapped_variables(variables, self.uri)

        return list_rows(
            uri=self.uri,
            max_results=self.properties["numRows"] if max_results is None else min(max_results,
                                                                                   int(self.properties["numRows"])),
            selected_variables=variables,
            mapped_variables=mapped_variables,
            output_type="polars_lazyframe",
            progress=progress,
            coerce_schema=hasattr(self.properties, "container") is False or self.properties["container"][
                "kind"] == 'dataset',
            batch_preprocessor=batch_preprocessor
        )

    def to_dask_dataframe(self, max_results=None, *, variables=None, progress=True, batch_preprocessor=None):
        if not self.properties or not hasattr(self.properties, "numRows"):
            self.get()

        mapped_variables = get_mapped_variables(variables, self.uri)

        return list_rows(
            uri=self.uri,
            max_results=self.properties["numRows"] if max_results is None else min(max_results,
                                                                                   int(self.properties["numRows"])),
            selected_variables=variables,
            mapped_variables=mapped_variables,
            output_type="dask_dataframe",
            progress=progress,
            coerce_schema=hasattr(self.properties, "container") is False or self.properties["container"][
                "kind"] == 'dataset',
            batch_preprocessor=batch_preprocessor
        )

    def to_pandas_dataframe(self, max_results=None, *, variables=None, progress=True, dtype_backend='pyarrow',
                            date_as_object=False, batch_preprocessor=None):
        if dtype_backend not in ['numpy', 'numpy_nullable', 'pyarrow']:
            raise Exception(
                f"Unknown dtype_backend. Must be one of 'pyarrow'|'numpy_nullable'|'numpy'. Default is 'pyarrow'")

        if not self.properties or not hasattr(self.properties, "numRows"):
            self.get()

        mapped_variables = get_mapped_variables(variables, self.uri)
        arrow_table = list_rows(
            uri=self.uri,
            max_results=self.properties["numRows"] if max_results is None else min(max_results,
                                                                                   int(self.properties["numRows"])),
            selected_variables=variables,
            mapped_variables=mapped_variables,
            output_type="arrow_table",
            progress=progress,
            coerce_schema=hasattr(self.properties, "container") is False or self.properties["container"][
                "kind"] == 'dataset',
            batch_preprocessor=batch_preprocessor
        )

        if dtype_backend == 'numpy_nullable':
            df = arrow_table.to_pandas(self_destruct=True, date_as_object=date_as_object, types_mapper={
                pa.int64(): pd.Int64Dtype(),
                pa.bool_(): pd.BooleanDtype(),
                pa.float64(): pd.Float64Dtype(),
                pa.string(): pd.StringDtype(),
            }.get)
        elif dtype_backend == 'pyarrow':
            df = arrow_table.to_pandas(self_destruct=True, types_mapper=pd.ArrowDtype)
        else:
            df = arrow_table.to_pandas(self_destruct=True, date_as_object=date_as_object)

        return df

    def to_geopandas_dataframe(self, max_results=None, *, variables=None, geography_variable="", progress=True,
                               dtype_backend='pyarrow', date_as_object=False, batch_preprocessor=None):
        import geopandas

        if dtype_backend not in ['numpy', 'numpy_nullable', 'pyarrow']:
            raise Exception(
                f"Unknown dtype_backend. Must be one of 'pyarrow'|'numpy_nullable'|'numpy'. Default is 'pyarrow'")

        if not self.properties or not hasattr(self.properties, "numRows"):
            self.get()

        mapped_variables = get_mapped_variables(variables, self.uri)
        arrow_table = list_rows(
            uri=self.uri,
            max_results=self.properties["numRows"] if max_results is None else min(max_results,
                                                                                   int(self.properties["numRows"])),
            selected_variables=variables,
            mapped_variables=mapped_variables,
            output_type="arrow_table",
            progress=progress,
            coerce_schema=hasattr(self.properties, "container") is False or self.properties["container"][
                "kind"] == 'dataset',
            batch_preprocessor=batch_preprocessor
        )

        if dtype_backend == 'numpy_nullable':
            df = arrow_table.to_pandas(self_destruct=True, date_as_object=date_as_object, types_mapper={
                pa.int64(): pd.Int64Dtype(),
                pa.bool_(): pd.BooleanDtype(),
                pa.float64(): pd.Float64Dtype(),
                pa.string(): pd.StringDtype(),
            }.get)
        elif dtype_backend == 'pyarrow':
            df = arrow_table.to_pandas(self_destruct=True, types_mapper=pd.ArrowDtype)
        else:
            df = arrow_table.to_pandas(self_destruct=True, date_as_object=date_as_object)

        if geography_variable is not None:
            geography_variable = get_geography_variable(mapped_variables, geography_variable)
            if geography_variable is None:
                raise Exception('Unable to find a variable with type=="geography" in the query results')

        if geography_variable is not None:
            df[geography_variable["name"]] = geopandas.GeoSeries.from_wkt(df[geography_variable["name"]])
            df = geopandas.GeoDataFrame(data=df, geometry=geography_variable["name"], crs="EPSG:4326")

        return df

    def to_dataframe(self, max_results=None, *, variables=None, geography_variable="", progress=True):
        warnings.warn(get_warning('dataframe_deprecation'), FutureWarning, stacklevel=2)

        if not self.properties or not hasattr(self.properties, "numRows"):
            self.get()

        mapped_variables = get_mapped_variables(variables, self.uri)
        arrow_table = list_rows(
            uri=self.uri,
            max_results=self.properties["numRows"] if max_results is None else min(max_results,
                                                                                   int(self.properties["numRows"])),
            selected_variables=variables,
            mapped_variables=mapped_variables,
            output_type="arrow_table",
            progress=progress,
            coerce_schema=hasattr(self.properties, "container") is False or self.properties["container"][
                "kind"] == 'dataset'
        )

        df = arrow_table.to_pandas(self_destruct=True)

        if geography_variable is not None:
            geography_variable = get_geography_variable(mapped_variables, geography_variable)

        if geography_variable is not None:
            import geopandas
            warnings.warn(get_warning('geodataframe_deprecation'), FutureWarning, stacklevel=2)
            df[geography_variable["name"]] = geopandas.GeoSeries.from_wkt(df[geography_variable["name"]])
            df = geopandas.GeoDataFrame(data=df, geometry=geography_variable["name"], crs="EPSG:4326")

        return df

    def to_arrow_batch_iterator(self, max_results=None, *, variables=None, progress=True):
        if not self.properties or not hasattr(self.properties, "numRows"):
            self.get()

        mapped_variables = get_mapped_variables(variables, self.uri)
        return list_rows(
            uri=self.uri,
            max_results=self.properties["numRows"] if max_results is None else min(max_results, int(self.properties["numRows"])),
            selected_variables=variables,
            mapped_variables=mapped_variables,
            output_type="arrow_iterator",
            progress=progress,
            coerce_schema=hasattr(self.properties, "container") is False or self.properties["container"][
                "kind"] == 'dataset'
        )

    def list_rows(self, max_results=None, *, variables=None, progress=True):
        warnings.warn(
            "The list_rows method is deprecated. Please use table.to_arrow_batch_iterator() or table.to_arrow_table().to_pylist() for better performance and memory utilization.",
            FutureWarning, stacklevel=2)

        if not self.properties or not hasattr(self.properties, "numRows"):
            self.get()

        mapped_variables = get_mapped_variables(variables, self.uri)
        return list_rows(
            uri=self.uri,
            max_results=self.properties["numRows"] if max_results is None else min(max_results,
                                                                                   int(self.properties["numRows"])),
            selected_variables=variables,
            mapped_variables=mapped_variables,
            output_type="tuple",
            progress=progress,
            coerce_schema=hasattr(self.properties, "container") is False or self.properties["container"][
                "kind"] == 'dataset'
        )

    def variable(self, name):
        return Variable(name, upload=self)


def upload_file(data, table_uri):
    start_byte = 0
    retry_count = 0
    chunk_size = MAX_CHUNK_SIZE
    did_reopen_file = False
    is_file = True if hasattr(data, "read") else False
    file_size = os.stat(data.name).st_size if is_file else len(data)

    if (is_file and hasattr(data, 'mode') and 'b' not in data.mode):
        data = open(data.name, 'rb')
        did_reopen_file = True

    res = make_request(
        path=f"{table_uri}/resumableUpload",
        method="POST",
        payload={"size": file_size},
    )
    resumable_uri = res["url"]
    resumable_upload_id = res["id"]

    while start_byte < file_size:
        end_byte = min(start_byte + chunk_size - 1, file_size - 1)
        if is_file:
            data.seek(start_byte)
            chunk = data.read(end_byte - start_byte + 1)
        else:
            chunk = data[start_byte : end_byte + 1]

        try:
            res = requests.put(
                url=resumable_uri,
                headers={
                    "Content-Length": f"{end_byte - start_byte + 1}",
                    "Content-Range": f"bytes {start_byte}-{end_byte}/{file_size}",
                },
                data=chunk,
            )
            res.raise_for_status()

            start_byte += chunk_size
            retry_count = 0
        except Exception as e:
            if retry_count > 20:
                print("A network error occurred. Upload failed after too many retries.")
                raise e

            retry_count += 1
            time.sleep(retry_count)
            print("A network error occurred. Retrying last chunk of resumable upload.")
            start_byte = retry_partial_upload(
                file_size=file_size, resumable_uri=resumable_uri
            )

    if did_reopen_file:
        data.close()

    return resumable_upload_id

def retry_partial_upload(*, retry_count=0, file_size, resumable_uri):
    logging.debug("Attempting to resume upload")

    try:
        res = requests.put(
            url=resumable_uri,
            headers={"Content-Length": "0", "Content-Range": f"bytes */{file_size}"},
        )

        if res.status_code == 404:
            return 0

        res.raise_for_status()

        if res.status_code == 200 or res.status_code == 201:
            return file_size
        elif res.status_code == 308:
            range_header = res.headers["Range"]

            if range_header:
                match = re.match(r"bytes=0-(\d+)", range_header)
                if match.group(0) and not math.isnan(int(match.group(1))):
                    return int(match.group(1)) + 1
                else:
                    raise Exception("An unknown error occurred. Please try again.")
            # If GCS hasn't received any bytes, the header will be missing
            else:
                return 0
    except Exception as e:
        if retry_count > 10:
            raise e

        time.sleep(retry_count / 10)
        retry_partial_upload(
            retry_count=retry_count + 1,
            file_size=file_size,
            resumable_uri=resumable_uri,
        )


def get_mapped_variables(variables, uri):
    all_variables = make_paginated_request(path=f"{uri}/variables", page_size=1000)

    if variables is None:
        return all_variables
    else:
        lower_variable_names = [variable.lower() for variable in variables]
        variables_list = list(
            filter(
                lambda variable: variable["name"].lower() in lower_variable_names,
                all_variables,
            )
        )
        variables_list.sort(
            key=lambda variable: lower_variable_names.index(variable["name"].lower())
        )
        return variables_list



from .Table import Table
from .Base import Base
from ..common.api_request import make_request
import pathlib
import uuid
import os
import pyarrow as pa
import pyarrow.dataset as pa_dataset
import pyarrow.parquet as pa_parquet
import pandas as pd

class Notebook(Base):
    def __init__(
        self,
        current_notebook_job_id,
    ):
        self.current_notebook_job_id = current_notebook_job_id

    def create_output_table(self, data=None, *, name=None, append=False, geography_variables=None):
        if type(data) == str:
            temp_file_path = str
        else:
            temp_file_path = f"/tmp/redivis/out/{uuid.uuid4()}"
            pathlib.Path(temp_file_path).parent.mkdir(exist_ok=True, parents=True)
            
            import geopandas

            if isinstance(data, geopandas.GeoDataFrame):
                if geography_variables is None:
                    geography_variables = list(data.select_dtypes('geometry'))
                data.to_wkt().to_parquet(path=temp_file_path, index=False)
            elif isinstance(data, pd.DataFrame):
                data.to_parquet(path=temp_file_path, index=False)
            elif isinstance(data, pa_dataset.Dataset):
                pa_dataset.write_dataset(data, temp_file_path, format='parquet', basename_template='part-{i}.parquet', max_partitions=1)
                temp_file_path = f'{temp_file_path}/part-0.parquet'
            elif isinstance(data, pa.Table):
                pa_parquet.write_table(data, temp_file_path)
            else:
                # importing polars is causing an IllegalInstruction error on ARM + Docker. Import inline to avoid crashes elsewhwere
                # TODO: revert once fixed upstream
                import polars
                if isinstance(data, polars.LazyFrame):
                    data.sink_parquet(temp_file_path)
                elif isinstance(data, polars.DataFrame):
                    data.write_parquet(temp_file_path)
                else:
                    raise Exception('Unknown datatype provided to notebook.create_output_table. Must either by a file path, or an instance of pandas.DataFrame, pyarrow.Dataset, pyarrow.Table, dask.DataFrame, polars.LazyFrame, or polars.DataFrame')

        with open(temp_file_path, 'rb') as f:
            res = make_request(
                method="PUT",
                path=f"/notebookJobs/{self.current_notebook_job_id}/outputTable",
                query={"name": name, "append": append, "geographyVariables": geography_variables},
                payload=f,
                parse_payload=False
            )

        if type(data) != str:
            os.remove(temp_file_path)

        return Table(name=res["name"], properties=res)



from .Base import Base
from urllib.parse import quote as quote_uri
import warnings
import os
import concurrent.futures
import pathlib
import pyarrow as pa
import pandas as pd
from tqdm.auto import tqdm

from .Upload import Upload
from .Variable import Variable
from .File import File
from ..common.list_rows import list_rows
from ..common.api_request import make_request, make_paginated_request
from ..common.util import get_geography_variable, get_warning


class Table(Base):
    def __init__(
        self,
        name,
        *,
        sample=False,
        dataset=None,
        project=None,
        properties={},
    ):
        dataset, project = get_table_parents(dataset, project)
        parent = dataset or project
        owner = parent.user or parent.organization
        sample_string = ":sample" if sample else ""
        version_string = f":{dataset.version}" if dataset else ""
        self.name = name
        self.dataset = dataset
        self.project = project
        self.identifier = (
            f"{owner.name}.{parent.name}{version_string}.{self.name}{sample_string}"
        )
        self.uri = f"/tables/{quote_uri(self.identifier, '')}"
        self.properties = {
            **{
                "kind": "table",
                "name": name,
                "uri": self.uri
            },
            **properties
        }

    def create(self, *, description=None, upload_merge_strategy="append", is_file_index=False):
        response = make_request(
            method="POST",
            path=f"{self.dataset.uri}/tables",
            payload={
                "name": self.name,
                "description": description,
                "uploadMergeStrategy": upload_merge_strategy,
                "isFileIndex": is_file_index
            },
        )
        self.properties = response
        self.uri = self.properties["uri"]
        return self

    def delete(self):
        make_request(
            method="DELETE",
            path=self.uri,
        )
        return

    def get(self):
        self.properties = make_request(method="GET", path=self.uri)
        self.uri = self.properties["uri"]
        return self

    def exists(self):
        try:
            make_request(method="GET", path=self.uri)
            return True
        except Exception as err:
            if err.args[0]["status"] != 404:
                raise err
            return False

    def list_uploads(self, *, max_results=None):
        uploads = make_paginated_request(
            path=f"{self.uri}/uploads", max_results=max_results
        )
        return [Upload(upload) for upload in uploads]

    def list_variables(self, *, max_results=None):
        variables = make_paginated_request(
            path=f"{self.uri}/variables", page_size=1000, max_results=max_results
        )
        return [
            Variable(variable["name"], table=self, properties=variable)
            for variable in variables
        ]

    def add_file(self, name, data):
        response = make_request(
            method="POST",
            path=f"{self.uri}/rawFiles",
            query={"name": name},
            payload=data,
            parse_payload=False
        )
        return File(response["id"], table=self, properties=response)

    def list_files(self, max_results=None, *, file_id_variable=None):
        if file_id_variable:
            variable = Variable(file_id_variable, table=self)
            if not variable.get().properties['isFileId']:
                raise Exception(f'The variable {file_id_variable} does not represent a file id.')
        else:
            variables = make_paginated_request(
                path=f"{self.uri}/variables", max_results=2, query={"isFileId": True}
            )
            if len(variables) == 0:
                raise Exception(f"No variable containing file ids was found on this table")
            elif len(variables) > 1:
                raise Exception(f"This table contains multiple variables representing a file id. Please specify the variable with file ids you want to download via the 'file_id_variable' parameter.")

            file_id_variable = variables[0]["name"]

        rows = self.to_arrow_table(max_results=max_results, variables=[file_id_variable], progress=False).to_pylist()
        return [
            File(row[file_id_variable], table=self)
            for row in rows
        ]

    def download_files(self, path=None, *, overwrite=False, max_results=None, file_id_variable=None, progress=True):
        files = self.list_files(max_results, file_id_variable=file_id_variable)
        if path is None:
            path = os.getcwd()

        if progress:
            pbar_count = tqdm(total=len(files), leave=False, unit=' files')
            pbar_bytes = tqdm(unit='B', leave=False, unit_scale=True, bar_format='{n_fmt} ({rate_fmt})')

        if not os.path.exists(path):
            pathlib.Path(path).mkdir(exist_ok=True, parents=True)


        def on_progress(bytes):
            nonlocal pbar_bytes
            pbar_bytes.update(bytes)

        def download(file):
            nonlocal progress
            nonlocal pbar_count

            file.download(path, overwrite=overwrite, progress=False, on_progress=on_progress if progress else None)
            if progress:
                pbar_count.update()

        with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
            try:
                executor.map(download, files)
            except KeyboardInterrupt:
                executor.shutdown(wait=False, cancel_futures=True)

        if progress:
            pbar_count.close()
            pbar_bytes.close()

    def to_arrow_dataset(self, max_results=None, *, variables=None, progress=True, batch_preprocessor=None):
        if not self.properties or not hasattr(self.properties, "numRows"):
            self.get()

        mapped_variables = get_mapped_variables(variables, self.uri)

        return list_rows(
            uri=self.uri,
            max_results=self.properties["numRows"] if max_results is None else min(max_results, int(self.properties["numRows"])),
            selected_variables=variables,
            mapped_variables=mapped_variables,
            output_type="arrow_dataset",
            progress=progress,
            coerce_schema=hasattr(self.properties, "container") is False or self.properties["container"]["kind"] == 'dataset',
            batch_preprocessor=batch_preprocessor
        )

    def to_arrow_table(self, max_results=None, *, variables=None, progress=True, batch_preprocessor=None):
        if not self.properties or not hasattr(self.properties, "numRows"):
            self.get()

        mapped_variables = get_mapped_variables(variables, self.uri)

        return list_rows(
            uri=self.uri,
            max_results=self.properties["numRows"] if max_results is None else min(max_results, int(self.properties["numRows"])),
            selected_variables=variables,
            mapped_variables=mapped_variables,
            output_type="arrow_table",
            progress=progress,
            coerce_schema=hasattr(self.properties, "container") is False or self.properties["container"]["kind"] == 'dataset',
            batch_preprocessor=batch_preprocessor
        )

    def to_polars_lazyframe(self, max_results=None, *, variables=None, progress=True, batch_preprocessor=None):
        if not self.properties or not hasattr(self.properties, "numRows"):
            self.get()

        mapped_variables = get_mapped_variables(variables, self.uri)

        return list_rows(
            uri=self.uri,
            max_results=self.properties["numRows"] if max_results is None else min(max_results, int(self.properties["numRows"])),
            selected_variables=variables,
            mapped_variables=mapped_variables,
            output_type="polars_lazyframe",
            progress=progress,
            coerce_schema=hasattr(self.properties, "container") is False or self.properties["container"]["kind"] == 'dataset',
            batch_preprocessor=batch_preprocessor
        )

    def to_dask_dataframe(self, max_results=None, *, variables=None, progress=True, batch_preprocessor=None):
        if not self.properties or not hasattr(self.properties, "numRows"):
            self.get()

        mapped_variables = get_mapped_variables(variables, self.uri)

        return list_rows(
            uri=self.uri,
            max_results=self.properties["numRows"] if max_results is None else min(max_results, int(self.properties["numRows"])),
            selected_variables=variables,
            mapped_variables=mapped_variables,
            output_type="dask_dataframe",
            progress=progress,
            coerce_schema=hasattr(self.properties, "container") is False or self.properties["container"]["kind"] == 'dataset',
            batch_preprocessor=batch_preprocessor
        )

    def to_pandas_dataframe(self, max_results=None, *, variables=None, progress=True, dtype_backend='pyarrow', date_as_object=False, batch_preprocessor=None):
        if dtype_backend not in ['numpy', 'numpy_nullable', 'pyarrow']:
            raise Exception(f"Unknown dtype_backend. Must be one of 'pyarrow'|'numpy_nullable'|'numpy'. Default is 'pyarrow'")

        if not self.properties or not hasattr(self.properties, "numRows"):
            self.get()

        mapped_variables = get_mapped_variables(variables, self.uri)
        arrow_table = list_rows(
            uri=self.uri,
            max_results=self.properties["numRows"] if max_results is None else min(max_results, int(self.properties["numRows"])),
            selected_variables=variables,
            mapped_variables=mapped_variables,
            output_type="arrow_table",
            progress=progress,
            coerce_schema=hasattr(self.properties, "container") is False or self.properties["container"][
                "kind"] == 'dataset',
            batch_preprocessor=batch_preprocessor
        )

        if dtype_backend == 'numpy_nullable':
            df = arrow_table.to_pandas(self_destruct=True, date_as_object=date_as_object, types_mapper={
                pa.int64(): pd.Int64Dtype(),
                pa.bool_(): pd.BooleanDtype(),
                pa.float64(): pd.Float64Dtype(),
                pa.string(): pd.StringDtype(),
            }.get)
        elif dtype_backend == 'pyarrow':
            df = arrow_table.to_pandas(self_destruct=True, types_mapper=pd.ArrowDtype)
        else:
            df = arrow_table.to_pandas(self_destruct=True, date_as_object=date_as_object)

        return df

    def to_geopandas_dataframe(self, max_results=None, *, variables=None, geography_variable="", progress=True, dtype_backend='pyarrow', date_as_object=False, batch_preprocessor=None):
        import geopandas

        if dtype_backend not in ['numpy', 'numpy_nullable', 'pyarrow']:
            raise Exception(f"Unknown dtype_backend. Must be one of 'pyarrow'|'numpy_nullable'|'numpy'. Default is 'pyarrow'")

        if not self.properties or not hasattr(self.properties, "numRows"):
            self.get()

        mapped_variables = get_mapped_variables(variables, self.uri)
        arrow_table = list_rows(
            uri=self.uri,
            max_results=self.properties["numRows"] if max_results is None else min(max_results, int(self.properties["numRows"])),
            selected_variables=variables,
            mapped_variables=mapped_variables,
            output_type="arrow_table",
            progress=progress,
            coerce_schema=hasattr(self.properties, "container") is False or self.properties["container"][
                "kind"] == 'dataset',
            batch_preprocessor=batch_preprocessor
        )

        if dtype_backend == 'numpy_nullable':
            df = arrow_table.to_pandas(self_destruct=True, date_as_object=date_as_object, types_mapper={
                pa.int64(): pd.Int64Dtype(),
                pa.bool_(): pd.BooleanDtype(),
                pa.float64(): pd.Float64Dtype(),
                pa.string(): pd.StringDtype(),
            }.get)
        elif dtype_backend == 'pyarrow':
            df = arrow_table.to_pandas(self_destruct=True, types_mapper=pd.ArrowDtype)
        else:
            df = arrow_table.to_pandas(self_destruct=True, date_as_object=date_as_object)

        if geography_variable is not None:
            geography_variable = get_geography_variable(mapped_variables, geography_variable)
            if geography_variable is None:
                raise Exception('Unable to find a variable with type=="geography" in the query results')

        if geography_variable is not None:
            df[geography_variable["name"]] = geopandas.GeoSeries.from_wkt(df[geography_variable["name"]])
            df = geopandas.GeoDataFrame(data=df, geometry=geography_variable["name"], crs="EPSG:4326")

        return df

    def to_dataframe(self, max_results=None, *, variables=None, geography_variable="", progress=True):
        warnings.warn(get_warning('dataframe_deprecation'), FutureWarning, stacklevel=2)

        if not self.properties or not hasattr(self.properties, "numRows"):
            self.get()

        mapped_variables = get_mapped_variables(variables, self.uri)
        arrow_table = list_rows(
            uri=self.uri,
            max_results=self.properties["numRows"] if max_results is None else min(max_results,
                                                                                   int(self.properties["numRows"])),
            selected_variables=variables,
            mapped_variables=mapped_variables,
            output_type="arrow_table",
            progress=progress,
            coerce_schema=hasattr(self.properties, "container") is False or self.properties["container"][
                "kind"] == 'dataset'
        )

        df = arrow_table.to_pandas(self_destruct=True)

        if geography_variable is not None:
            geography_variable = get_geography_variable(mapped_variables, geography_variable)

        if geography_variable is not None:
            import geopandas
            warnings.warn(get_warning('geodataframe_deprecation'), FutureWarning, stacklevel=2)
            df[geography_variable["name"]] = geopandas.GeoSeries.from_wkt(df[geography_variable["name"]])
            df = geopandas.GeoDataFrame(data=df, geometry=geography_variable["name"], crs="EPSG:4326")

        return df

    def to_arrow_batch_iterator(self, max_results=None, *, variables=None, progress=True):
        if not self.properties or not hasattr(self.properties, "numRows"):
            self.get()

        mapped_variables = get_mapped_variables(variables, self.uri)
        return list_rows(
            uri=self.uri,
            max_results=self.properties["numRows"] if max_results is None else min(max_results, int(self.properties["numRows"])),
            selected_variables=variables,
            mapped_variables=mapped_variables,
            output_type="arrow_iterator",
            progress=progress,
            coerce_schema=hasattr(self.properties, "container") is False or self.properties["container"][
                "kind"] == 'dataset'
        )

    def list_rows(self, max_results=None, *, variables=None, progress=True):
        warnings.warn("The list_rows method is deprecated. Please use table.to_arrow_batch_iterator() or table.to_arrow_table().to_pylist() for better performance and memory utilization.", FutureWarning, stacklevel=2)

        if not self.properties or not hasattr(self.properties, "numRows"):
            self.get()

        mapped_variables = get_mapped_variables(variables, self.uri)
        return list_rows(
            uri=self.uri,
            max_results=self.properties["numRows"] if max_results is None else min(max_results, int(self.properties["numRows"])),
            selected_variables=variables,
            mapped_variables=mapped_variables,
            output_type="tuple",
            progress=progress,
            coerce_schema=hasattr(self.properties, "container") is False or self.properties["container"]["kind"] == 'dataset'
        )

    def update(self, *, name=None, description=None, upload_merge_strategy=None):
        payload = {}
        if name:
            payload["name"] = name
        if upload_merge_strategy:
            payload["uploadMergeStrategy"] = upload_merge_strategy
        if description is not None:
            payload["description"] = description

        response = make_request(
            method="PATCH",
            path=f"{self.uri}",
            payload=payload,
        )
        self.properties = response
        return

    def upload(
        self,
        name,
        *,
        type=None,
        schema=None,
        has_header_row=True,
        skip_bad_records=False,
        allow_quoted_newlines=False,
        quote_character='"',
        delimiter=None,
        #     This is deprecated
        data=None,
        remove_on_fail=True,
    ):
        upload = Upload(
            table=self,
            name=name,
        )
        if data is not None:
            warnings.warn(
                "Passing data directly to the upload constructor is deprecated. Please call table.upload('filename').create(data) instead.",
                FutureWarning,
                stacklevel=2
            )
            upload.create(
                schema=schema,
                type=type,
                has_header_row=has_header_row,
                skip_bad_records=skip_bad_records,
                has_quoted_newlines=allow_quoted_newlines,
                quote_character=quote_character,
                delimiter=delimiter,
            )
            upload.upload_file(
                data,
                wait_for_finish=True,
                raise_on_fail=True,
                remove_on_fail=remove_on_fail,
            )

        return upload

    def variable(self, name):
        return Variable(name, table=self)


def get_mapped_variables(variables, uri):
    all_variables = make_paginated_request(path=f"{uri}/variables", page_size=1000)
    if variables is None:
        return all_variables
    else:
        lower_variable_names = [variable.lower() for variable in variables]
        variables_list = list(
            filter(
                lambda variable: variable["name"].lower() in lower_variable_names,
                all_variables,
            )
        )
        variables_list.sort(
            key=lambda variable: lower_variable_names.index(variable["name"].lower())
        )
        return variables_list


def get_table_parents(dataset, project):
    from .User import User

    if dataset or project:
        return dataset, project
    elif os.getenv("REDIVIS_DEFAULT_PROJECT") is not None:
        return None, User(os.getenv("REDIVIS_DEFAULT_PROJECT").split(".")[0]).project(
            os.getenv("REDIVIS_DEFAULT_PROJECT").split(".")[1]
        )
    elif os.getenv("REDIVIS_DEFAULT_DATASET") is not None:
        return (
            User(os.getenv("REDIVIS_DEFAULT_DATASET").split(".")[0]).dataset(
                os.getenv("REDIVIS_DEFAULT_DATASET").split(".")[1]
            )
        ), None
    else:
        raise Exception(
            "Cannot reference an unqualified table if the neither the REDIVIS_DEFAULT_PROJECT or REDIVIS_DEFAULT_DATASET environment variables are set."
        )
